/*	motion.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/videodev.h>

#include <jpeglib.h>

#define DEF_WIDTH	352
#define DEF_HEIGHT	288
#define DEF_QUALITY	50
#define DEF_CHANGES	1500

#define DEF_GAP		60*5 /* 5 minutes */

#define NORM_DEFAULT	0
#define NORM_PAL	0
#define NORM_NTSC	1
#define NORM_SECAM	2
#define IN_DEFAULT	8
#define IN_TV		0
#define IN_COMPOSITE	1
#define IN_COMPOSITE2	2
#define IN_SVIDEO	3

#define VIDEO_DEVICE		"/dev/video0"

int deamon=0;
int snapshot=0;
int alarmtime=0;

void usage (void)
{
	printf("motion Version "VERSION", Copyright 2000 Jeroen Vreeken\n");
	printf("usage:\n");
	printf("\n");
	printf("motion [options]\n");
	printf("\n");
	printf("-a seconds\ttime between two automated snapshots, default: 0\n");
	printf("-c changes\tthreshold for detecting a change, default: %d\n", DEF_CHANGES);
	printf("-D\t\tDeamonize\n");
	printf("-d device\tvideo4linux capture device, default: "VIDEO_DEVICE"\n");
	printf("-g seconds\tminimum gap between movements for mail/sms messages, default %d\n", DEF_GAP);
	printf("-i input\tinput channel to use, default: not set\n");
	printf("-M address\tSend a mail to 'address' when detecting motion.\n");
	printf("-n norm\t\tnorm to use (pal/ntsc/secam), default: pal\n");
	printf("-q quality\tJPEG image quality, default: %d\n", DEF_QUALITY);
	printf("-S nr\t\tSend a SMS to nr using sms_client when detecting motion.\n");
	printf("-s widthxheight\tPicture size, default: %dx%d\n", DEF_WIDTH, DEF_HEIGHT);
	printf("-t target dir\tdestination for snapshots.\n");
	printf("\n");
}

char *start_capture (int dev, int width, int height, int input, int norm,  int fmt)
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
	struct video_mbuf vid_buf;
	char *map;

	if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (NULL);
	}
	/* Set channels and norms, NOT TESTED my philips cam doesn't have a
	 * tuner. */
        if (input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev, VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = input;
                        vid_chnl.norm    = norm;
                        if (ioctl (dev, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (NULL);
                        }
                }
        }
	if (ioctl (dev, VIDIOCGMBUF, &vid_buf) == -1)
		return (NULL);
	map=mmap(0, vid_buf.size, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	if ((unsigned char *)-1 == (unsigned char *)map)
		return (NULL);
	return map;
}

char *next_capture (int dev, char *map, int width, int height, int fmt)
{
	struct video_mmap vid_mmap;

	vid_mmap.format=fmt;
	vid_mmap.frame=0;
	vid_mmap.width=width;
	vid_mmap.height=height;
	if (ioctl(dev, VIDIOCMCAPTURE, &vid_mmap) == -1)
		return (NULL);
	if (ioctl(dev, VIDIOCSYNC, &vid_mmap) == -1)
		return (NULL);
	return map;
}

void put_image_jpeg (FILE *picture, char *image, int width, int height, int quality)
{
        int y, x, line_width;
        JSAMPROW row_ptr[1];
        struct jpeg_compress_struct cjpeg;
        struct jpeg_error_mgr jerr;
        char *line;

        line = malloc (width * 3);
        if (!line)
                return;
        cjpeg.err = jpeg_std_error(&jerr);
        jpeg_create_compress (&cjpeg);
        cjpeg.image_width = width;
        cjpeg.image_height= height;
        cjpeg.input_components = 3;
        cjpeg.in_color_space = JCS_RGB;
        jpeg_set_defaults (&cjpeg);

        jpeg_set_quality (&cjpeg, quality, TRUE);
        cjpeg.dct_method = JDCT_FASTEST;
        jpeg_stdio_dest (&cjpeg, picture);

        jpeg_start_compress (&cjpeg, TRUE);

        row_ptr[0] = line;
        line_width = width * 3;
	for ( y = 0; y < height; y++) {
        for (x = 0; x < line_width; x+=3) {
                        line[x]   = image[x+2];
                        line[x+1] = image[x+1];
                        line[x+2] = image[x];
                }
                jpeg_write_scanlines (&cjpeg, row_ptr, 1);
                image += line_width;
        }
        jpeg_finish_compress (&cjpeg);
        jpeg_destroy_compress (&cjpeg);
        free (line);
}

void send_sms(char *sms_nr, struct tm *currenttime)
{
	char message[150];

	if (!fork()) {
	    	printf("Sending SMS message to %s\n", sms_nr);
		sprintf(message, "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		execlp("sms_client", "sms_client", sms_nr, message, 0);
		perror("Unable to send SMS message");
		exit(0);
	}
	return;
}

void send_mail(char *mail_address, struct tm *currenttime)
{
	char message[150];
	int fdpipe[2];
	
	if (!fork()) {
	    	printf("Sending mail to %s\n", mail_address);
		sprintf(message, "This is an automated message generated by motion\n\n"
		    "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n"
		    ".\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		pipe(fdpipe);
		if (!deamon) {
			close(0);
			dup(fdpipe[0]);
			close(fdpipe[0]);
		}
		write(fdpipe[1], message, strlen(message));
		close(fdpipe[1]);
		execlp("mail", "mail", mail_address, "-s", "MOTION ALERT", 0);
		perror("Unable to send message");
		exit(0);
	}
	return;
}

void sig_alarm (int signo) 
{
	/* Somebody (maybe we ourself) wants us to make a snapshot */
	snapshot=1;
	/* Set everything for the next snapshot */
	signal(SIGALRM, sig_alarm);
	if (alarmtime) alarm(alarmtime); 
}

int main (int argc, char **argv)
{
	int width=DEF_WIDTH;
	int height=DEF_HEIGHT;
	int quality=DEF_QUALITY;
	int max_changes=DEF_CHANGES;
	int dev, i, diffs, detected=0, shots=0, sms=0, mail=0, gap=DEF_GAP;
	int input=IN_DEFAULT, norm=0;
	unsigned char *image_ref, *image_new, *image_out;
	time_t currenttimep, lasttime=0;
	struct tm *currenttime=NULL;
	char *device=VIDEO_DEVICE, *mail_address=NULL, *sms_nr=NULL;
	char file[255], *filepath=NULL, c;
	FILE *picture;

	while ((c=getopt(argc, argv, "a:c:d:Dg:hi:M:n:q:S:s:t:"))!=EOF) {
		switch (c) {
			case 'a':
				sscanf (optarg, "%d", &alarmtime);
				break;
			case 'c':
				sscanf (optarg, "%d", &max_changes);
				break;
			case 'd':
				device=optarg;
				break;
			case 'D':
				deamon=1;
				break;
			case 'g':
				sscanf (optarg, "%d", &gap);
				break;
			case 'i':
				sscanf (optarg, "%d", &input);
				break;
			case 'M':
				mail=1;
				mail_address=optarg;
				break;
			case 'n':
				if (!strcmp(optarg, "pal")) norm=NORM_PAL;
				if (!strcmp(optarg, "ntsc")) norm=NORM_NTSC;
				if (!strcmp(optarg, "secam")) norm=NORM_SECAM;
				break;
			case 'S':
				sms=1;
				sms_nr=optarg;
				break;
			case 'q':
				sscanf (optarg, "%d", &quality);
				break;
			case 's':
				sscanf (optarg, "%dx%d", &width, &height);
				break;
			case 't':
				filepath=optarg;
				break;
			case 'h':
			/* Do the same as an invallid option, but without the
			 * 'invalid option' error.... */
			default:
				usage();
				exit(-1);
		}
	}
	/* Set signal handlers */
	signal(SIGCHLD, SIG_IGN);
	signal(SIGALRM, sig_alarm);
	if (alarmtime) alarm(alarmtime);

	image_ref=malloc(width*height*3);
	image_out=malloc(width*height*3);

	dev=open (device, O_RDWR);
	if (dev < 0) {
		perror ("Failed to open video device");
		exit(-1);
	}

	/* Deamonize */
	if (deamon) {
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}
	/* Capture first image, or we will get a alarm on start */
	image_new=start_capture (dev, width, height, input, norm, VIDEO_PALETTE_RGB24);
	/* Allow videodevice to settle in */
	sleep(1);
	next_capture(dev, image_new, width, height, VIDEO_PALETTE_RGB24);
	if (!image_new)
		perror("capture error");
	memcpy(image_ref, image_new, width*height*3);

	/* Should go on forever... unless you bought vaporware :) */
	while (next_capture(dev, image_new, width, height, VIDEO_PALETTE_RGB24)) {
		/* Make a differences picture in image_out */
		diffs=0;
		for (i=0; i<width*height*3; i++) {
			image_out[i]=image_ref[i]-image_new[i];
			if ((signed char)image_out[i] > 32 || (signed char)image_out[i] < -32) {
				image_out[i]=image_new[i];
				diffs++;
			} else {
				image_out[i]=0;
			}
		}
		currenttimep=time(NULL);
		currenttime=localtime(&currenttimep);
		/* The cat just walked in :) */
		if (diffs > max_changes) {
			detected=1;
			printf("\007Motion detected %d\n", diffs);
		}
		if (diffs > max_changes || snapshot) {
			if (lasttime!=currenttimep) {
				shots=0;
			}
			if (filepath) {
				sprintf(file, "%s/%04d%02d%02d%02d%02d%02d-%02d.jpg",
				    filepath,
				    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
				    currenttime->tm_mday, currenttime->tm_hour,
				    currenttime->tm_min, currenttime->tm_sec, shots);
			} else {
				sprintf(file, "%04d%02d%02d%02d%02d%02d-%02d.jpg",
				    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
				    currenttime->tm_mday, currenttime->tm_hour,
				    currenttime->tm_min, currenttime->tm_sec, shots);
			}
			printf("Picture saved to: %s\n", file);
			picture=fopen(file, "w");
			/* Output the latest picture 'image_new'.
			 * replace it with 'image_out' for neat effects, but less
			 * recognizable
			 */
			put_image_jpeg (picture, image_new, width, height, quality);
			fclose(picture);
			snapshot=0;
		}
		if (diffs > max_changes) {
			if (currenttimep-lasttime>= gap) {
				if (sms) send_sms(sms_nr, currenttime);
				if (mail) send_mail(mail_address, currenttime);
			}
			lasttime=currenttimep;
			shots++;
		} else {
			if (detected && mail && currenttimep-lasttime>= gap) {
				printf("should attach pictures to a mail\n");
				detected=0;
			}
		}
		/* Use this to minimize the number of pictures, also reduces
		 * the chance of a good one... However it can cause some
		 * seconds delay due to buffering in the v4l drivers...
		 */
//		sleep(1);
		/* Old image slowly decays, this will make it even harder on
		 * slow moving object to stay undetected
		 */
		for (i=0; i<width*height*3; i++) {
			image_ref[i]=(image_ref[i]+image_new[i])/2;
		}
	}
	perror("Somebody stole the video device, lets hope we got his picture");
	close (dev);
	free(image_ref);
	free(image_out);
	exit(0);
}